/* $Id: randr.c,v 1.2 2005/12/20 23:39:57 cegger Exp $
***************************************************************************

   X server for LibGGI - RANDR extension

   Copyright (C) 2003	Christoph Egger

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/


#include <X11/X.h>
#include "os.h"
#include "mibank.h"
#include "mipointer.h"
#include <randrstr.h>

#include "xggi.h"
#include "xggi_Xext.h"



typedef struct _RandRInfo {
	CreateScreenResourcesProcPtr CreateScreenResources;
	CloseScreenProcPtr CloseScreen;

	int virtualX;
	int virtualY;
} RandRInfoRec, *RandRInfoPtr;

static int RandRIndex;
static unsigned int RandRGeneration;

#define RANDRINFO(p)    ((RandRInfoPtr) (p)->devPrivates[RandRIndex].ptr)


static int RandRModeRefresh (ggi_mode *mode)
{
	return 60;
}	/* RandRModeRefresh */


static Bool RandRGetInfo(ScreenPtr pScreen, Rotation * rotations)
{
	int scr;
	RRScreenSizePtr pSize;
	RandRInfoPtr    randrp = RANDRINFO(pScreen);
	ggi_mode	*mode;
	int refresh0 = 60;
	int refresh;

	scr = xggiGetScreenIdx(pScreen);
	mode = &MASTER_VISUAL_MODE(scr);

	*rotations = RR_Rotate_0;

#if 0
	/* Bail if no depth has a visual associated with it */
	for (n = 0; n < pScreen->numDepths; n++) {
		if (pScreen->allowedDepths[n].numVids) {
			break;
		}
	}
	if (n == pScreen->numDepths) {
		return FALSE;
	}
#endif

	refresh = RandRModeRefresh (mode);
	pSize = RRRegisterSize(pScreen,
				(short)mode->virt.x, (short)mode->virt.y,
				(short)pScreen->mmWidth, (short)pScreen->mmHeight);
	if (!pSize) return FALSE;

	RRRegisterRate(pScreen, pSize, refresh);
	RRSetCurrentConfig(pScreen, RR_Rotate_0, refresh, pSize);

	if ((mode->visible.x != randrp->virtualX)
	   || (mode->visible.y != randrp->virtualY))
	{
		pSize = RRRegisterSize(pScreen,
				       randrp->virtualX, randrp->virtualY,
				       pScreen->mmWidth * randrp->virtualX / mode->visible.x,
				       pScreen->mmHeight * randrp->virtualY / mode->visible.y);

		if (!pSize) return FALSE;
		RRRegisterRate(pScreen, pSize, refresh0);

		if ((pScreen->width == randrp->virtualX) &&
		    (pScreen->height == randrp->virtualY))
		{
			RRSetCurrentConfig(pScreen, RR_Rotate_0, refresh0, pSize);
		}	/* if */
	}	/* if */

	return TRUE;
}	/* RandRGetInfo */



static Bool RandRSetMode(int scr,
			ggi_mode *mode,
			Bool useVirtual)
{
	ScreenPtr	pScreen = SCREEN_SCR(scr);
	RandRInfoPtr    randrp = RANDRINFO(pScreen);
	int             oldWidth = pScreen->width;
	int             oldHeight = pScreen->height;

#if 0
	xf86EnableDisableFBAccess(pScreen->myNum, FALSE);
#endif

	if (useVirtual) {
		mode->virt.x = randrp->virtualX;
		mode->virt.y = randrp->virtualY;
	} else {
		mode->virt.x = mode->visible.x;
		mode->virt.y = mode->visible.y;
	}	/* if */

	pScreen->width = mode->virt.x;
	pScreen->height = mode->virt.y;
	if (!xggiSwitchMode(scr, mode)) {
		mode->virt.x = pScreen->width = oldWidth;
		mode->virt.y = pScreen->height = oldHeight;
		return FALSE;
	}	/* if */

#if 0
	/*
	 * Make sure the layout is correct
	 */
	xf86ReconfigureLayout();
#endif

#if 0
	/*
	 * Make sure the whole screen is visible
	 */
	xf86SetViewport(pScreen, pScreen->width, pScreen->height);
	xf86SetViewport(pScreen, 0, 0);
#endif
#if 0
	xf86EnableDisableFBAccess(pScreen->myNum, TRUE);
#endif

	return TRUE;
}	/* RandRSetMode */


/*
 * Implements RRSetConfigPtr() function hook
 */
static Bool RandRSetConfig(ScreenPtr pScreen,
		       Rotation rotation,
		       int rate,
		       RRScreenSizePtr pSize)
{
	int scr;
	RandRInfoPtr    randrp = RANDRINFO(pScreen);
	ggi_mode	*mode;
	int             px, py;
	Bool            useVirtual = FALSE;

	scr = xggiGetScreenIdx(pScreen);
	mode = &MASTER_VISUAL_MODE(scr);

	miPointerPosition(&px, &py);
	if ((mode->visible.x == pSize->width) &&
	    (mode->visible.y == pSize->height) &&
	    ((rate == 0) || (RandRModeRefresh(mode) == rate)))
	{
		goto further;
	}

	if ((pSize->width == randrp->virtualX) &&
	    (pSize->height == randrp->virtualY))
	{
		/* mode pointer already initialized */
#if 0
		mode = ;
#endif
		useVirtual = TRUE;
		goto further;
	}	/* if */
	return FALSE;

further:
	if (!RandRSetMode(scr, mode, useVirtual)) {
		return FALSE;
	}	/* if */

	/*
	 * Move the cursor back where it belongs; SwitchMode repositions it
	 */
	if (pScreen == miPointerCurrentScreen()) {
		if (px < pSize->width && py < pSize->height) {
			(*pScreen->SetCursorPosition) (pScreen, px, py, FALSE);
		}	/* if */
	}	/* if */
	return TRUE;
}	/* RandRSetConfig */



/*
 * Wait until the screen is initialized before whacking the
 * sizes around; otherwise the screen pixmap will be allocated
 * at the current mode size rather than the maximum size
 */
static Bool RandRCreateScreenResources(ScreenPtr pScreen)
{
	RandRInfoPtr randrp = RANDRINFO(pScreen);

	pScreen->CreateScreenResources = randrp->CreateScreenResources;
	if (!(*pScreen->CreateScreenResources) (pScreen)) {
		return FALSE;
	}	/* if */

	return TRUE;
}	/* RandRCreateScreenResources */


/*
 * Reset size back to original
 */
static Bool RandRCloseScreen(int idx, ScreenPtr pScreen)
{
	int scr = xggiGetScreenIdx(pScreen);
	ggi_mode	*mode = &MASTER_VISUAL_MODE(scr);
	RandRInfoPtr    randrp = RANDRINFO(pScreen);

	mode->virt.x = pScreen->width = randrp->virtualX;
	mode->virt.y = pScreen->height = randrp->virtualY;
	pScreen->CloseScreen = randrp->CloseScreen;
	xfree(randrp);
	pScreen->devPrivates[RandRIndex].ptr = 0;
	return (*pScreen->CloseScreen) (idx, pScreen);
}	/* RandRCloseScreen */




Bool xggiXextRandRInit(int scr)
{
	rrScrPrivPtr    rp;
	RandRInfoPtr    randrp;
	ggi_mode	*mode;
	ScreenPtr	pScreen = SCREEN_SCR(scr);

#ifdef PANORAMIX
	/* XXX disable RandR when using Xinerama */
	if (!noPanoramiXExtension)
		return TRUE;
#endif
	if (RandRGeneration != serverGeneration) {
		RandRIndex = AllocateScreenPrivateIndex();
		RandRGeneration = serverGeneration;
	}	/* if */
	randrp = xalloc(sizeof(RandRInfoRec));
	if (!randrp)
		return FALSE;

	if (!RRScreenInit(pScreen)) {
		xfree(randrp);
		return FALSE;
	}	/* if */

	mode = &MASTER_VISUAL_MODE(scr);

	rp = rrGetScrPriv(pScreen);
	rp->rrGetInfo = RandRGetInfo;
	rp->rrSetConfig = RandRSetConfig;

	randrp->virtualX = mode->virt.x;
	randrp->virtualY = mode->virt.y;

	randrp->CreateScreenResources = pScreen->CreateScreenResources;
	pScreen->CreateScreenResources = RandRCreateScreenResources;

	randrp->CloseScreen = pScreen->CloseScreen;
	pScreen->CloseScreen = RandRCloseScreen;

	pScreen->devPrivates[RandRIndex].ptr = randrp;
	return TRUE;
}	/* xggiXextRandRInit */
