/* $Id: gcfuncs.c,v 1.2 2006/01/21 16:17:14 cegger Exp $
***************************************************************************

   X server for LibGGI - GC

   Copyright (C) 1997	   Jason McMullan	[jmcc@cs.cmu.edu]
   Copyright (C) 1997	   Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998-2000 Marcus Sundberg	[marcus@ggi-project.org]
   Copyright (C) 2003-2006 Christoph Egger

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/


#include "gc.h"


static void
xggiValidateGC(GCPtr pGC, unsigned long changes, DrawablePtr pDrawable);

static void xggiDestroyGC(GCPtr pGC);

static void
xggiChangeGC(GCPtr pGC, unsigned long mask);

static void
xggiCopyGC(GCPtr pGCSrc, unsigned long mask,
	GCPtr pGCDst);

static void
xggiChangeClip(GCPtr pGC, int type, pointer pvalue, int nrects);

static void
xggiCopyClip(GCPtr pgcDst, GCPtr pgcSrc);

static void
xggiDestroyClip(GCPtr pGC);




GCFuncs xggiGCFuncs = {
	xggiValidateGC, xggiChangeGC, xggiCopyGC, xggiDestroyGC,
	xggiChangeClip, xggiDestroyClip, xggiCopyClip
};

GCOps xggiGCOps = {
	xggiFillSpans, xggiSetSpans, xggiPutImage, xggiCopyArea,
	xggiCopyPlane, xggiPolyPoint, xggiPolylines, xggiPolySegment,
	xggiPolyRectangle, xggiPolyArc, xggiFillPolygon, xggiPolyFillRect,
	xggiPolyFillArc, xggiPolyText8, xggiPolyText16, xggiImageText8,
	xggiImageText16, xggiImageGlyphBlt, xggiPolyGlyphBlt,
	xggiPushPixels,
#ifdef NEED_LINEHELPER
	NULL,
#endif
	{ NULL }	/* devPrivate */
};



Bool xggiCreateGC(GCPtr pGC)
{
	Bool ret = TRUE;
	ScreenPtr pScreen = pGC->pScreen;
	xggiScreenBackupPtr pScrPriv;
	xggiGCPrivPtr pGCPriv;
	int scr;

	scr = xggiGetScreenIdx(pScreen);
	pScrPriv = SCREEN_SCRBACKUP(scr);
	pGCPriv = GCBACKUP(scr, pGC);

	unwrap (pScrPriv, pScreen, CreateGC);
	ret = (*pScreen->CreateGC)(pGC);
	if (ret) {
		pGCPriv->ops = NULL;	/* don't wrap ops yet */
		pGCPriv->funcs = pGC->funcs;
		pGC->funcs = &xggiGCFuncs;
	}
	wrap (pScrPriv, pScreen, CreateGC, xggiCreateGC);

	return ret;
}


#define XGGI_GC_FUNC_PROLOGUE(pGC)			\
	int scr = xggiGetScreenIdx(pGC->pScreen);	\
	xggiGCPrivPtr pGCPriv = GCBACKUP(scr, pGC);	\
	unwrap(pGCPriv, pGC, funcs);			\
	if (pGCPriv->ops) unwrap(pGCPriv, pGC, ops)

#define XGGI_GC_FUNC_EPILOGUE(pGC)			\
	wrap (pGCPriv, pGC, funcs, &xggiGCFuncs);	\
	if (pGCPriv->ops) wrap(pGCPriv, pGC, ops, &xggiGCOps)


static void
xggiValidateGC(GCPtr pGC, unsigned long changes, DrawablePtr pDrawable)
{
	XGGI_GC_FUNC_PROLOGUE(pGC);
	(*pGC->funcs->ValidateGC)(pGC, changes, pDrawable);
	pGCPriv->ops = pGC->ops;	/* let ops now wrap */
	XGGI_GC_FUNC_EPILOGUE(pGC);
}


static void xggiDestroyGC(GCPtr pGC)
{
	XGGI_GC_FUNC_PROLOGUE(pGC);
	(*pGC->funcs->DestroyGC)(pGC);
	XGGI_GC_FUNC_EPILOGUE(pGC);
}	/* xggiDestroyGC */


static void
xggiChangeGC(GCPtr pGC, unsigned long mask)
{
	XGGI_GC_FUNC_PROLOGUE(pGC);
	(*pGC->funcs->ChangeGC)(pGC, mask);
	XGGI_GC_FUNC_EPILOGUE(pGC);
}


static void
xggiCopyGC(GCPtr pGCSrc, unsigned long mask,
	GCPtr pGCDst)
{
	XGGI_GC_FUNC_PROLOGUE(pGCDst);
	(*pGCDst->funcs->CopyGC)(pGCSrc, mask, pGCDst);
	XGGI_GC_FUNC_EPILOGUE(pGCDst);
}


static void
xggiChangeClip(GCPtr pGC, int type, pointer pvalue, int nrects)
{
	XGGI_GC_FUNC_PROLOGUE(pGC);
	(*pGC->funcs->ChangeClip)(pGC, type, pvalue, nrects);
	XGGI_GC_FUNC_EPILOGUE(pGC);
}


static void
xggiCopyClip(GCPtr pgcDst, GCPtr pgcSrc)
{
	XGGI_GC_FUNC_PROLOGUE(pgcDst);
	(* pgcDst->funcs->CopyClip)(pgcDst, pgcSrc);
	XGGI_GC_FUNC_EPILOGUE(pgcDst);
}

static void
xggiDestroyClip(GCPtr pGC)
{
	XGGI_GC_FUNC_PROLOGUE(pGC);
	(*pGC->funcs->DestroyClip)(pGC);
	XGGI_GC_FUNC_EPILOGUE(pGC);
}

