/* $Id: mode.c,v 1.1.1.1 2005/12/18 17:38:09 cegger Exp $
***************************************************************************

   X server for LibGGI - Mode handling

   Copyright (C) 1997	   Jason McMullan	[jmcc@cs.cmu.edu]
   Copyright (C) 1997	   Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998-2000 Marcus Sundberg	[marcus@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/

#include "xggi.h"



int xggiSwitchMode(int scr, ggi_mode *mode)
{
	int j, size;
	ggi_graphtype gt;
	const ggi_directbuffer *dbuf = NULL;

	if (MASTER_VISUAL(scr).dbuf != NULL) {
		/* happens, when we switch to an other mode */
		ggiResourceRelease(MASTER_VISUAL(scr).dbuf->resource);
	}	/* if */

	ggiCheckMode(MASTER_VISUAL_VIS(scr), mode);
#if 0
	if (scr == 0) {
		gt = MASTER_VISUAL_MODE(scr).graphtype;
	}	/* if */
#endif
	gt = MASTER_VISUAL_MODE(MASTER_SCREEN_NR).graphtype;

	if (GT_SIZE(mode->graphtype) != 1 &&
	    GT_SIZE(mode->graphtype) != 8 &&
	    GT_SIZE(mode->graphtype) != 16 &&
	    GT_SIZE(mode->graphtype) != 24 &&
	    GT_SIZE(mode->graphtype) != 32)
	{
		FatalError("XGGI doesn't support %d bpp screens.\n",
			   GT_SIZE(mode->graphtype));
	}	/* if */
	if (mode->graphtype != gt) {
		/* All GGI visuals must be of the same depth. */
		return -30;
	}	/* if */

	MASTER_VISUAL_WIDTH(scr) = mode->visible.x;
	MASTER_VISUAL_HEIGHT(scr) = mode->visible.y;
	MASTER_VISUAL_DEPTH(scr) = GT_DEPTH(mode->graphtype);
	MASTER_VISUAL_BPP(scr) = GT_SIZE(mode->graphtype);

	if (ggiSetMode(MASTER_VISUAL_VIS(scr), mode) < 0) {
		/* LibGGI can not set any modes at all! */
		return -31;
	}	/* if */

	size = GT_SIZE(mode->graphtype);
	for (j = 0;
	    (dbuf = ggiDBGetBuffer(MASTER_VISUAL_VIS(scr), j)) != NULL;
	     j++)
	{
#ifdef XGGI_DEBUG
		ErrorF("InitGGI: directbuffer type: %d\n", dbuf->type);
#endif
		if ((dbuf->type & GGI_DB_SIMPLE_PLB)
		    && ((8*dbuf->buffer.plb.stride) % size) == 0)
		{
			MASTER_VISUAL(scr).dbuf = dbuf;
			break;
		}	/* if */
	}	/* for */

	if (!MASTER_VISUAL(scr).dbuf) {
		/* This mode and target has no suitable DirectBuffer */
		FatalError("This mode and target has no suitable DirectBuffer\n");
		return -32;
	}	/* if */

	if (ggiResourceAcquire(MASTER_VISUAL(scr).dbuf->resource,
			       GGI_ACTYPE_WRITE | GGI_ACTYPE_READ)
	    != 0)
	{
		FatalError("Unable to acquire DirectBuffer\n");
		return -33;
	}	/* if */

	MASTER_VISUAL(scr).pfbMemory = dbuf->write;
	MASTER_VISUAL_STRIDE(scr) = 8 * dbuf->buffer.plb.stride / size;
	MASTER_VISUAL(scr).dbuf = dbuf;

	return TRUE;
}	/* xggiSwitchMode */
